from starlette.requests import Request

from .base import MiddlewareLimiter


class IPLimiter(MiddlewareLimiter):
    def __call__(self, request: Request):
        host = request.headers.get(
            "X_FORWARDED_FOR", request.client.host or "127.0.0.1"
        )
        self.limit(key=f"host={host}")


class TotalLimiter(MiddlewareLimiter):
    def __call__(self, request: Request):
        self.limit(key="total")



"""
This module defines in-memory storage for WindowManager's
"""


import time
from typing import MutableMapping, Optional

from throttled.models import Hit, Rate
from throttled.storage import BaseStorage
from throttled.storage._abstract import _HitsWindow, _WindowManager
from throttled.storage._duration import DUR_REGISTRY, DurationCalcType
from throttled.strategies import Strategies


class _MemoryWindow(_HitsWindow):
    __slots__ = ("__expire_at", "__hits")

    def __init__(self, duration: float):
        self.__expire_at = time.time() + duration
        self.__hits = 0

    def incr(self, hits: int = 1) -> int:
        self.__hits += hits
        return self.__hits

    def decr(self, hits: int = 1) -> int:
        """
        To be implemented:

        self.__hits -= hits
        return self.__hits
        """

    def get_remaining_seconds(self) -> float:
        return self.__expire_at - time.time()


class _MemoryWindowManager(_WindowManager):
    __slots__ = ("__interval", "__cache", "__duration_calc")

    def __init__(
        self,
        interval: float,
        duration_func: DurationCalcType,
        cache: Optional[MutableMapping[str, _MemoryWindow]] = None,
    ):
        self.__cache: MutableMapping[str, _MemoryWindow] = (
            cache if cache is not None else {}
        )
        self.__interval = interval
        self.__duration_calc = duration_func

    def get_current_window(self, hit: Hit) -> _MemoryWindow:
        window = self.__cache.get(hit.key)
        if window is None or window.is_expired():
            self.__cache[hit.key] = window = _MemoryWindow(
                duration=self.__duration_calc(hit.time, self.__interval)
            )
        return window


class MemoryStorage(BaseStorage):
    def __init__(
        self,
        cache: Optional[MutableMapping[str, _MemoryWindow]] = None,
    ):
        self.__cache = cache

    def get_window_manager(
        self, strategy: Strategies, limit: Rate
    ) -> _MemoryWindowManager:
        return _MemoryWindowManager(
            interval=limit.interval,
            duration_func=DUR_REGISTRY[strategy],
            cache=self.__cache,
        )


__all__ = ["MemoryStorage"]



import time as clock
from dataclasses import dataclass, field
from typing import NamedTuple


@dataclass(frozen=True)
class Hit:
    key: str = field(
        default_factory=str,
        metadata={"description": "The base contribution to the hit key."},
    )
    time: float = field(
        default_factory=clock.time,
        metadata={
            "units": "seconds",
            "description": "time the hit occurred since epoch.",
        },
    )
    cost: int = 1


class Rate(NamedTuple):
    hits: int
    interval: float



from typing import Callable, Mapping

from throttled.strategies import Strategies

DurationCalcType = Callable[[float, float], float]


def fixed_window_duration_calc(hit_time: float, interval: float) -> float:
    return interval - hit_time % interval


def moving_window_duration_calc(_: float, interval: float) -> float:
    return interval


DUR_REGISTRY: Mapping[Strategies, DurationCalcType] = {
    Strategies.FIXED_WINDOW: fixed_window_duration_calc,
    Strategies.MOVING_WINDOW: moving_window_duration_calc,
}



from abc import ABC, abstractmethod
from typing import Callable, Optional, Sequence

from fastapi.exceptions import HTTPException
from starlette import status
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response

from throttled.exceptions import RateLimitExceeded
from throttled.limiter import Limiter
from throttled.models import Rate
from throttled.storage import BaseStorage
from throttled.strategies import Strategies


class HTTPLimitExceeded(HTTPException):
    def __init__(self, exc: RateLimitExceeded, detail: Optional[str] = None):
        self.retry_after = exc.retry_after
        super_kwargs = dict(
            status_code=status.HTTP_429_TOO_MANY_REQUESTS,
            headers={"Retry-After": str(round(self.retry_after, 2))},
        )
        if detail is not None:
            super_kwargs["detail"] = detail
        super().__init__(**super_kwargs)


DetailFactory = Callable[[RateLimitExceeded], Optional[str]]


def key_detail_factory(exc: RateLimitExceeded) -> str:
    return f"Rate exceeded for key={exc.key}."


def null_detail_factory(_: RateLimitExceeded) -> None:
    return None


class FastAPILimiter(ABC):
    """First adapter between limiters APIs and FastAPI"""

    def __init__(
        self,
        limit: Rate,
        storage: BaseStorage,
        strategy: Strategies = Strategies.MOVING_WINDOW,
        detail_factory: DetailFactory = key_detail_factory,
    ):
        self.__limiter = Limiter(limit, storage, strategy)
        self.__detail_factory = detail_factory

    def limit(self, key: str):
        """
        Limit the hit based on it's key

        :param key: the key to be limited.
        :raises: HTTPLimitExceeded
        """
        try:
            self.__limiter.limit(key)
        except RateLimitExceeded as exc:
            raise HTTPLimitExceeded(exc, detail=self.__detail_factory(exc)) from exc

    @abstractmethod
    def __call__(self, request: Request):
        """
        This method implementation are supposed to call the limit() method with
        the key to be limited.

        :param request: injected by FastAPI when using the limiter as dependency.
        :raises: HTTPLimitExceeded
        """


ResponseFactory = Callable[[HTTPLimitExceeded], Response]


def default_response_factory(exc: HTTPLimitExceeded) -> Response:
    return Response(
        status_code=exc.status_code,
        headers=exc.headers,
    )


class MiddlewareLimiter(FastAPILimiter, ABC):
    __ignored_paths: Sequence[str] = ("docs", "redoc", "favicon.ico", "openapi.json")

    def __init__(
        self,
        limit: Rate,
        storage: BaseStorage,
        strategy: Strategies = Strategies.MOVING_WINDOW,
        detail_factory: DetailFactory = key_detail_factory,
        response_factory: ResponseFactory = default_response_factory,
    ):
        super().__init__(
            strategy=strategy,
            limit=limit,
            storage=storage,
            detail_factory=detail_factory,
        )
        self.__response_factory = response_factory

    def ignore_path(self, path: str):
        path = path.lstrip("/")
        ignored = list(self.__ignored_paths)
        ignored.append(path)
        self.__ignored_paths = ignored

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        """
        Dispatch function which converts the limiter in a Middleware.
        Middlewares are faster because they don't need dependencies to be resolved.

        All limiters that don't use FastAPI dependency injection can be implemented as Middleware.

        :param request: injected by FastAPI when using the limiter as middleware.
        :param call_next: the next middleware from the chain-of-responsibility created by Starlette.
        """
        try:
            path = str(request.url).replace(str(request.base_url), "")
            if path not in self.__ignored_paths:
                self(request)
            return await call_next(request)
        except HTTPLimitExceeded as exc:
            return self.__response_factory(exc)



from abc import ABC, abstractmethod

from throttled.models import Hit


class _HitsWindow(ABC):
    @abstractmethod
    def get_remaining_seconds(self) -> float:
        """Get the time (seconds) in which this HitsWindow will expire."""

    @abstractmethod
    def incr(self, hits: int = 1) -> int:
        """
        Increase the hit count in this window

        :param hits: the number of hits to increase in the window
        :return: the current hits number, after increased
        """

    @abstractmethod
    def decr(self, hits: int = 1) -> int:
        """
        Decrease the hit count in this window

        :param hits: the number of hits to decrease in the window
        :return: the current hits number, after decreased
        """

    def is_expired(self) -> bool:
        """
        Check if this window already expired

        :return: boolean True if the window already expired, False otherwise
        """
        return self.get_remaining_seconds() <= 0


class _WindowManager(ABC):
    @abstractmethod
    def get_current_window(self, hit: Hit) -> _HitsWindow:
        """
        Get the current active HitsWindow

        :param hit: hit to get the HitsWindow for
        :return: the currently active HitsWindow instance
        """



from typing import List, Tuple, Union

from fastapi import Depends
from starlette.middleware.base import DispatchFunction

from .base import FastAPILimiter, MiddlewareLimiter


def split_dependencies_and_middlewares(
    *limiters: Union[FastAPILimiter, MiddlewareLimiter]
) -> Tuple[List[Depends], List[DispatchFunction]]:

    dispatch_functions: List[DispatchFunction] = []
    dependencies: List[Depends] = []

    for limiter in limiters:
        if isinstance(limiter, MiddlewareLimiter):
            dispatch_functions.append(limiter.dispatch)
        elif isinstance(limiter, FastAPILimiter):
            dependencies.append(Depends(limiter))
        else:
            raise TypeError(f"Object {limiter} is not a Middleware or Limiter.")
    return dependencies, dispatch_functions



import asyncio
import functools
from typing import Any, Callable, TypeVar, cast

from throttled.exceptions import RateLimitExceeded
from throttled.models import Hit, Rate
from throttled.storage import BaseStorage
from throttled.strategies import Strategies

FuncT = TypeVar("FuncT", bound=Callable[..., Any])


class Limiter:
    __slots__ = ("__limit", "__window_manager")

    def __init__(
        self,
        limit: Rate,
        storage: BaseStorage,
        strategy: Strategies,
    ):
        self.__limit = limit
        self.__window_manager = storage.get_window_manager(
            strategy=strategy, limit=limit
        )

    def __maybe_block(self, hit: Hit):
        """
        :param hit: The hit to be tested
        :raises: RateLimitExceeded
        """
        window = self.__window_manager.get_current_window(hit)
        if window.incr(hit.cost) > self.__limit.hits:
            raise RateLimitExceeded(
                hit.key,
                retry_after=window.get_remaining_seconds(),
            )

    def limit(self, key: str):
        self.__maybe_block(Hit(key=key))

    def decorate(self, func: FuncT) -> FuncT:
        key = f"func={func.__name__}:hash={hash(func)}"

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self.limit(key)
            return func(*args, **kwargs)

        @functools.wraps(func)
        async def a_wrapper(*args, **kwargs):
            self.limit(key)
            return await func(*args, **kwargs)

        return cast(FuncT, a_wrapper if asyncio.iscoroutinefunction(func) else wrapper)



"""
This module defines redis storage for WindowManager's
"""

from redis.client import Redis

from throttled.models import Hit, Rate
from throttled.storage import BaseStorage
from throttled.storage._abstract import _HitsWindow, _WindowManager
from throttled.storage._duration import DUR_REGISTRY, DurationCalcType
from throttled.strategies import Strategies


class _RedisWindow(_HitsWindow):
    __slots__ = ("__client", "__interval", "__hit", "__duration_func")

    def __init__(
        self, interval: float, client: Redis, hit: Hit, duration_func: DurationCalcType
    ):
        self.__client = client
        self.__interval = interval
        self.__duration_func = duration_func
        self.__hit = hit

    def incr(self, hits: int = 1) -> int:
        value = self.__client.incrby(name=self.__hit.key, amount=hits)
        if value == hits:
            self.__client.pexpire(
                name=self.__hit.key,
                time=int(self.__duration_func(self.__hit.time, self.__interval) * 1e3),
            )
        return value

    def decr(self, hits: int = 1) -> int:
        """
        To be implemented:

        value = self.__client.decrby(name=self.__hit.key, amount=hits)
        return value
        """

    def get_remaining_seconds(self) -> float:
        return self.__client.pttl(name=self.__hit.key) * 1e-3


class _RedisWindowManager(_WindowManager):
    __slots__ = ("__interval", "__client", "__duration_func")

    def __init__(self, interval: float, client: Redis, duration_func: DurationCalcType):
        self.__client = client
        self.__interval = interval
        self.__duration_func = duration_func

    def get_current_window(self, hit: Hit) -> _RedisWindow:
        return _RedisWindow(
            client=self.__client,
            interval=self.__interval,
            hit=hit,
            duration_func=self.__duration_func,
        )


class RedisStorage(BaseStorage):
    def __init__(
        self,
        client: Redis,
    ):
        self.__client = client

    def get_window_manager(
        self, strategy: Strategies, limit: Rate
    ) -> _RedisWindowManager:
        return _RedisWindowManager(
            interval=limit.interval,
            duration_func=DUR_REGISTRY[strategy],
            client=self.__client,
        )


__all__ = ["RedisStorage"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['throttled', 'throttled.fastapi', 'throttled.storage']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'throttled',
    'version': '0.1.16a34',
    'description': 'A rate limiter for FastAPI',
    'long_description': '# ThrottledAPI\n\nThis repo aims to be an audacious RateLimiter for python ASGI/WSGI APIs.\n\n## RoadMap\n\n- [ ] Implement for FastAPI, because I really like the framework\n- [ ] Generalise for others\n\n## Status\n\nIn development ...\n',
    'author': 'Vinícius Vargas',
    'author_email': 'santunionivinicius@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/santunioni/ThrottledAPI',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod

from throttled.models import Rate
from throttled.storage._abstract import _WindowManager
from throttled.strategies import Strategies


class BaseStorage(ABC):
    @abstractmethod
    def get_window_manager(self, strategy: Strategies, limit: Rate) -> _WindowManager:
        """Return a _WindowManager for the given strategy and rate limit."""



from .base import FastAPILimiter, MiddlewareLimiter
from .limiters import IPLimiter, TotalLimiter
from .utils import split_dependencies_and_middlewares

__all__ = [
    "TotalLimiter",
    "IPLimiter",
    "split_dependencies_and_middlewares",
    "FastAPILimiter",
    "MiddlewareLimiter",
]
